### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Workers for sterch.conveyor package """

__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL"

import sys
import traceback
import types

from assistance import InQueueMixin, OutQueueMixin, EventMixin, LogMixin
from interfaces import IFirstWorker, ILastWorker, IRegularWorker
from Queue import Empty, Full
from threading import Thread
from time import sleep
from zope.component import getUtility
from zope.interface import implements

class FirstWorker(EventMixin, 
                  OutQueueMixin, 
                  LogMixin,
                  Thread):
    """ Initial worker """
    implements(IFirstWorker)
    
    def __init__(self, out_queue, event, delay, activity=None): 
        """    out_queue --- either object that provides IQueue or IQueue utility name
               event     --- either object that provides IEvent or IEvent utility name 
               delay     --- delay between activity cycles
               activity --- callable represents worker activity
        """
        Thread.__init__(self)
        self._out_queue = out_queue
        self.delay = delay
        self._event = event
        if activity:
            self.activity = activity
                
    def activity(self):
        raise NotImplementedError("Must be defined")
    
    def run(self):
        """ Worker's workcycle """
        while True:
            if self.event.isSet(): return
            try:
                for task in self.activity():
                    while True:
                        try:
                            self.out_queue.put(task, timeout=self.delay)
                            break
                        except Full, ex:
                            pass
            except Exception, ex:
                # Thread must stop if and only if event is set
                self.traceback(ex)

class LastWorker(EventMixin, 
                 InQueueMixin,
                 LogMixin,
                 Thread):
    """ Last worker """
    implements(ILastWorker)
    
    def __init__(self, in_queue, event, delay, activity=None): 
        """    in_queue --- either object that provides IQueue or IQueue utility name
               event     --- either object that provides IEvent or IEvent utility name 
               delay     --- delay between activity cycles
               activity --- callable represents worker activity
        """
        Thread.__init__(self)
        self._in_queue = in_queue
        self.delay = delay
        self._event = event
        if activity:  self.activity = activity
    
    def activity(self):
        raise NotImplementedError("Must be defined")
    
    def run(self):
        """ Worker's workcycle """
        while True:
            if self.event.isSet(): return
            try:
                try:
                    task = self.in_queue.get(timeout=self.delay)
                    self.activity(task)                        
                except Empty, ex:
                    pass
            except Exception, ex:
                # Thread must stop if and only if event is set
                self.traceback(ex)


class Worker(EventMixin, 
             InQueueMixin,
             OutQueueMixin,
             LogMixin,
             Thread):
    """ Regular worker to process elements from input queue to output queue.
        Could be stopped be setting event. 
    """
    implements(IRegularWorker)
    
    def __init__(self, in_queue, out_queue, event, delay, activity=None):
        """ 
            in_queue --- input queue
            out_queue --- output queue
            timeout --- time to wait in inqueue is empty and check evtAllDone
            event --- event to stop thread activity
            activity --- callable represents worker activity. Must accept only one argument --- item.
                        Returned value will be placed to out_queue if not None.
        """
        Thread.__init__(self)
        self._in_queue = in_queue
        self._out_queue = out_queue
        self.delay = delay
        self._event = event
        self.activity = activity
        if activity: self.activity = activity
        
    def activity(self, item):
        raise NotImplementedError("Must be implemented")
    
    def run(self):
        """ Worker cycle """
        while True:
            if self.event.isSet(): return
            try:
                try:
                    task = self.in_queue.get(timeout=self.delay)
                    for new_task in self.activity(task):
                        while True:
                            try:
                                self.out_queue.put(new_task, timeout=self.delay)
                                break
                            except Full, ex:
                                pass
                except Empty, ex:
                    pass
            except Exception, ex:
                # Thread must stop if and only if event is set
                self.traceback(ex)
                
RegularWorker = Worker


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2008
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2008
#######################################################################

"""Metadirectives implementations for the ZTK based sterch.threading package

"""
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL" 

from interfaces import IConveyor, IFirstWorker, ILastWorker, IRegularWorker
from zope.component import createObject, getUtility
from zope.component.interfaces import IFactory 
from zope.component.zcml import handler
from zope.configuration.exceptions import ConfigurationError

class ConveyorDirective(object):
    """ Conveyor directive handler """
    
    def __init__(self, _context, name=None, delay=5.0):
        self._context = _context
        self.name = name
        self.delay = delay
        self.init_stage_vars = False
        self.final_stage_vars = False
        self.stages = []
        self.all_in_queues = []
        
    def init_stage(self, _context, **kwargs):
        """ init_stage subdirective processing """
        kwargs['event'] = kwargs['event'] if kwargs.get('event') else createObject('threading.Event')
        kwargs['delay'] = kwargs['delay'] if kwargs.get('delay') else 0
        self.init_stage_vars = kwargs
        
        
    def final_stage(self, _context, **kwargs):
        """ final_stage subdirective processing """
        kwargs['event'] = kwargs['event'] if kwargs.get('event') else createObject('threading.Event')
        kwargs['delay'] = kwargs['delay'] if kwargs.get('delay') else 0
        self.final_stage_vars = kwargs
        self.all_in_queues.append(kwargs['in_queue'])
        
    def stage(self, _context, **kwargs):
        """ stage subdirective procesing """
        kwargs['event'] = kwargs['event'] if kwargs.get('event') else createObject('threading.Event')
        kwargs['delay'] = kwargs['delay'] if kwargs.get('delay') else 0
        self.stages.append(kwargs)
        self.all_in_queues.append(kwargs['in_queue'])
        
    def __call__(self):
        """ whole directive processing  """
        if not self.init_stage_vars: raise ConfigurationError(u"No initial stage defined")
        if not self.final_stage_vars: raise ConfigurationError(u"No final stage defined")
        # trying to build conveyor chain
        ordered_stages = [self.init_stage_vars]
        curq = self.init_stage_vars['out_queue']
        in_queues = []
        out_queues = [curq]
        while self.stages:
            if curq not in self.all_in_queues: raise ConfigurationError(u"Output queue %s is never used as input" % curqueue)
            s = self.stages.pop()
            if s['in_queue'] in in_queues: raise ConfigurationError(u"Input queue could be used only in one stage.")
            if s['out_queue'] in out_queues: raise ConfigurationError(u"Output queue could be used only in one stage.")
            if s['in_queue'] in in_queues: raise ConfigurationError(u"Input queue could be used only in one stage.")
            if s['out_queue'] in out_queues: raise ConfigurationError(u"Output queue could be used only in one stage.")
            if s.get('marker') == curq : raise ConfigurationError("Final stage unreachable")
            if s['in_queue'] == curq:
                ordered_stages.append(s)
                curq = s['out_queue']
                in_queues.append(s['in_queue'])
                out_queues.append(s['out_queue'])
            else:
                s['marker'] = curq
                self.stages = [s,] + self.stages
        if self.stages: ConfigurationError(u"There are dummy stages within the conveyor")
        if ordered_stages[-1]['out_queue'] != self.final_stage_vars['in_queue']:
            raise ConfigurationError(u"Conveyor must be finished with final_stage.")            
        ordered_stages.append(self.final_stage_vars)
        # removing markers
        for s in ordered_stages: 
            if 'marker' in s: del s['marker']
        # constructing the conveyor
        __stages = []
        # 1st stage
        s = ordered_stages[0]
        __stages.append(createObject('sterch.conveyor.FirstStage', **s))
        # middle stages
        for s in ordered_stages[1:-1]:
            __stages.append(createObject('sterch.conveyor.RegularStage', **s))
        #last stage
        s = ordered_stages[-1]
        __stages.append(createObject('sterch.conveyor.LastStage', **s))
        conveyor = createObject('sterch.conveyor.Conveyor', self.name, __stages, delay=self.delay)
        self._context.action(
            discriminator = ('utility', IConveyor, self.name),
            callable = handler,
            args = ('registerUtility', conveyor, IConveyor, self.name)
        )


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" data processing stage
"""
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL"

from assistance import InQueueMixin
from group import Group
from interfaces import IGroup, IRegularStage, IFirstStage, ILastStage
from zope.interface import implements
from zope.component import getUtility
from zope.component.interfaces import IFactory

class StageBase(object):
    """ Base data processing stage.
        This class DOES NOT implemet IStage.
        There is no way to implement has_tasks
    """
        
    def __init__(self, name, group):
        """  
            name --- stage name
            group --- worker's group
        """
        self._name = name
        self.group = group   
               
    @property
    def name(self):
        """ name must be read only according to IStage """
        return self._name
    
    def start(self):
        """ Start stage """
        self.group.start()
        
    def stop(self):
        """ Tries to stop stage """
        self.group.stop()
        
    def is_finished(self):
        """ Checks does group of workers finish its activity """
        self.group.filter_dead_workers()
        return len(self.group) == 0
    
    def has_tasks(self):
        """ Must be implemented in descendants """
        raise NotImplementedError("Not implemented yet.")
    
    def tasks_count(self):
        """ Must be implemented in descendants """
        raise NotImplementedError("Not implemented yet.")
    
    def workers_count(self):
       return self.group.live_count() 
    
class FirstStage(StageBase):
    """ First data processing stage. 
        Initial tasks generation.
    """
    implements(IFirstStage)
    
    def __init__(self, name, quantity, out_queue, event, delay, activity):
        """ 
            name --- stage name
            quantity --- number of workers to process stage
            out_queue --- queue to put tasks
            event --- event to stop all workers
            delay --- processing delay
            activity --- callable represents stage activity 
        """
        w_factory = getUtility(IFactory, name="sterch.conveyor.FirstWorker") 
        g_factory = getUtility(IFactory, name="sterch.conveyor.Group")
        w_args = (out_queue, event, delay, activity)
        group = g_factory(quantity, w_factory, *w_args)
        StageBase.__init__(self, name, group)
        
    def has_tasks(self):
        """ Initial stage has no unfinished tasks."""
        return False
    
    def tasks_count(self):
        """ Initial stage has no unfinished tasks."""
        return 0
    
class LastStage(InQueueMixin, StageBase):
    """ Last data processing stage. 
        Results generation.
    """
    implements(ILastStage)
    
    def __init__(self, name, quantity, in_queue, event, delay, activity):
        """ 
            name --- stage name
            quantity --- number of workers to process stage
            in_queue --- queue to get tasks
            event --- event to stop all workers
            delay --- processing delay
            activity --- callable represents stage activity 
        """
        self._in_queue = in_queue
        w_factory = getUtility(IFactory, name="sterch.conveyor.LastWorker") 
        g_factory = getUtility(IFactory, name="sterch.conveyor.Group")
        w_args = (in_queue, event, delay, activity)
        group = g_factory(quantity, w_factory, *w_args)
        StageBase.__init__(self, name, group)

class RegularStage(InQueueMixin, StageBase):
    """ Regular data processing stage. 
        Take tasks from input queue and put next tasks to output queue.
    """
    implements(IRegularStage)
    def __init__(self, name, quantity, in_queue, out_queue, event, delay, activity):
        """ 
            name --- stage name
            quantity --- number of workers to process stage
            in_queue --- queue to get tasks
            out_queue --- queue to put next tasks
            event --- event to stop all workers
            delay --- processing delay
            activity --- callable represents stage activity 
        """
        self._in_queue = in_queue
        w_factory = getUtility(IFactory, name="sterch.conveyor.RegularWorker") 
        g_factory = getUtility(IFactory, name="sterch.conveyor.Group")
        w_args = (in_queue, out_queue, event, delay, activity)
        group = g_factory(quantity, w_factory, *w_args)
        StageBase.__init__(self, name, group)


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Worker's group
"""
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL"


from interfaces import IGroup, IWorker
from sterch.threading.interfaces import IThread
from threading import Lock 
from UserList import UserList
from zope.interface import implements 

class Group(UserList):
    """ Thread-safe group of workers """

    def __init__(self, qty, factory, *args, **kwargs):
        """ Creates group of workers to make similar work.
            qty --- number of workers
            factory --- workers' factory
            *args, **kwargs --- args for an factory
        """
        super(Group, self).__init__()
        for i in xrange(0,qty):
            # this will guarantee that all list elements are workers
            self.append(IWorker(factory(*args, **kwargs)))
        self.locked = Lock()
    
    def start(self):
        """ start all workers """
        with self.locked:
            map(lambda t:IThread(t).start(), self)
    
    def stop(self):
        """ Fire event to stop all workers """
        with self.locked:
            map(lambda t:IWorker(t).event.set(), self)
            
    def is_finished(self):
        """ Check workers to be stopped """
        return self.live_count() == 0
    
    def filter_dead_workers(self):
        """ remove dead workers from the group """
        with self.locked:
            self.data = filter(lambda t: IThread(t).isAlive(), self)
    
    def live_count(self):
        """ returns number of live workers """
        with self.locked:
            return len(filter(lambda t: IThread(t).isAlive(), self))


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Base test classes for sterch.conveyor
"""
__author__  = "Maxim Polscha (maxp@sterch.net)"
__license__ = "ZPL" 

import sterch.logfile
import sterch.queue
import sterch.threading
import zope.app.component
from unittest import TestCase, makeSuite, main
from zope.component.testing import PlacelessSetup
from zope.configuration.xmlconfig import XMLConfig 

class TestSetup(PlacelessSetup, TestCase):
    """Test the various zcml configurations"""
    
    def setUp(self):
        super(TestSetup, self).setUp()
        XMLConfig('meta.zcml', zope.app.component)()
        XMLConfig('meta.zcml', sterch.threading)()
        XMLConfig('meta.zcml', sterch.queue)()
        XMLConfig('meta.zcml', sterch.logfile)()
        XMLConfig('meta.zcml', sterch.conveyor)()
        XMLConfig('configure.zcml', sterch.threading)()
        XMLConfig('configure.zcml', sterch.queue)()
        XMLConfig('configure.zcml', sterch.logfile)() 
        XMLConfig('configure.zcml', sterch.conveyor)()


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Base test classes for sterch.conveyor
"""
__author__  = "Maxim Polscha (maxp@sterch.net)"
__license__ = "ZPL" 

import zcml

from setup import TestSetup
from unittest import TestCase, makeSuite, main
from zope.configuration.xmlconfig import XMLConfig
from zope.configuration.exceptions import ConfigurationError

class Test(TestSetup):
    """Test the various zcml configurations"""
    
    def test_loop(self):
        self.assertRaises(ConfigurationError, XMLConfig, *('loop.zcml', zcml)) 

    def test_simple_loop(self):
        self.assertRaises(ConfigurationError, XMLConfig, *('simple_loop.zcml', zcml)) 
        
    def test_no_init_stage(self):
        self.assertRaises(ConfigurationError, XMLConfig, *('no_init_stage.zcml', zcml)) 
                
    def test_no_final_stage(self):
        self.assertRaises(ConfigurationError, XMLConfig, *('no_final_stage.zcml', zcml)) 
        
    def test_final_stage_unreachable(self):
        self.assertRaises(ConfigurationError, XMLConfig, *('final_stage_unreachable.zcml', zcml)) 
        
    def test_dummy_stages(self):
        self.assertRaises(ConfigurationError, XMLConfig,* ('dummy_stages.zcml', zcml)) 

def test_suite():
    suite = makeSuite(Test)
    return suite

if __name__ == '__main__':
    main(defaultTest='test_suite')        


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################
"""Interfaces for the ZTK based sterch.worker package

"""
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL" 

from sterch.queue.interfaces import IQueue
from sterch.threading.interfaces import IThread, IEvent
from zope.interface import Interface
from zope.interface.common.sequence import ISequence
from zope.schema import Object, Int, Float, Tuple, TextLine

class IEventMixin(Interface):
    event = Object(title=u"Event", schema=IEvent, required=True, readonly=True)
    
class IDelayMixin(Interface):
    delay = Float(title=u"Delay in sec. between activity cycles", min=0.0, default=0.0, required=True, readonly=True)
    
class IOutQueueMixin(Interface):
    out_queue = Object(title=u"Queue to put tasks", schema=IQueue, readonly=True, required=True)

class IInQueueMixin(Interface):
    in_queue = Object(title=u"Queue to get tasks", schema=IQueue, readonly=True, required=True)
    
class IWorker(IThread,
                  IEventMixin,
                  IDelayMixin):
    """ Wroker base interface """
    event = Object(title=u"Event to stop worker", schema=IEvent, required=True, readonly=True)
    
    def activity(*args, **kwargs):
        """ Main worker activity. 
            MUST be generator.  """

class IFirstWorker(IWorker, IOutQueueMixin):
    """ First worker that creates initial tasks queue """
    
    def activity():
        """ Activity have no input args. This is initial tasks generator. """
    
class ILastWorker(IWorker, IInQueueMixin):
    """ Last worker that processes last tasks que queue """

class IRegularWorker(IWorker, IInQueueMixin, IOutQueueMixin):
    """ Regular worker. It gets tasks from input queue and puts to output queue """

class IDelayedFinish(Interface):
    """ Interface of delayed finish.
        It describes situation when stopping is parallel task 
        and may take a lot of time.
    """
    def stop():
        """ Do something to stop (f.e. fire event) """
        
    def is_finished():
        """ Check was it really finished """
        
class IGroup(ISequence, IDelayedFinish):
    """ Sequence of workers """
    name = TextLine(title=u"Group name", required=True, readonly=True)
    
    def start():
        """" start all workers """
    
    def filter_dead_workers():
        """ removes dead workers from sequence """
        
    def live_count():
        """ number of active workers """

class IStage(IDelayedFinish, IEventMixin, IDelayMixin):
    """ Data processing stage """
    name = TextLine(title=u"Stage name", required=True, readonly=True)
    
    def has_tasks():
        """ Returns True if stage has unprocessed tasks finished, False otherwise. """

    def tasks_count():
        """ Returns number of unprocessed tasks. """
        
    def workers_count():
        """ Returns number of live workers. """
    
    def start():
        """ starts stage """
            
class IFirstStage(IStage, IOutQueueMixin):
    """ Initial data processing stage """

class ILastStage(IStage, IInQueueMixin):
    """ Final data processing stage """

class IRegularStage(IStage, IInQueueMixin, IOutQueueMixin):
    """ Regular data processing stage. Requires both input and output queues """
        
class IConveyor(IThread, IDelayMixin):
    """ Conveyor of parallel data processing.
        When starts it MUST start all stages' workers
    """
    name = TextLine(title=u"Conveyor name", required=True, readonly=True)
    stages = Tuple(title=u"All stages ordered.", readonly=True, required=True)


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Conveyor for the sterch.conveyor package """
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL"

from assistance import LogMixin
from interfaces import IConveyor, IStage
from threading import Thread
from time import sleep 
from zope.cachedescriptors.property import CachedProperty
from zope.interface import implements

class Conveyor(Thread, LogMixin):
    """ Simple IConveyor implementation """
    implements(IConveyor)
    
    def __init__(self, name, stages, delay=5):
        """ delay --- delay in secs. between stages check
            stages --- ordered list of stages to process data """
        Thread.__init__(self)
        self.name = name
        self._delay = delay
        self._stages = [IStage(s) for s in stages]

    @CachedProperty
    def stages(self): return self._stages
    
    @CachedProperty
    def delay(self): return self._delay
    
    def run(self):
        """ main conveyor activity """
        # start all stages
        self.log.message("Starting all stages.")
        map(lambda s:s.start(), self.stages)
        
        for stage in self.stages:
            # Wait for queue
            while stage.has_tasks():
                self.log.message("Waiting for stage %s. %s tasks queued." % (stage.name, stage.tasks_count()))
                sleep(self.delay)
            stage.stop()
            self.log.message("Stage %s. All tasks were completed." % stage.name)
            # wait for  workers
            while not stage.is_finished() :
                self.log.message("Stage %s. Waiting for workers. There are %s active workers." % (stage.name, stage.workers_count()))
                sleep(self.delay)
            self.log.message("Stage %s. All workers were stopped." % stage.name)
        self.log.message("Done.")


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

"""ZCML directives interfaces for the ZTK based sterch.conveyor package

"""
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL"

from sterch.threading.metadirectives import IName
from zope.interface import Interface, invariant, Invalid
from zope.configuration.fields import GlobalObject
from zope.schema import TextLine, Int, Float
       
class IConveyorDirective(IName):
    """ <conveyor ...>  complex directive interface """
    delay = Float(title=u"Delay to wait stages in queue (sec., 5 by default)", min=0.0, required=False, default=5.0)
    
class IStageBase(IName):
    """ generic stage fields """
    activity = GlobalObject(title=u"Callable object to process tasks on the stage.",
                            constraint=lambda obj:hasattr(obj,'__call__'))
    quantity = Int(title=u"Number of workers for the stage.", min=0, required=True)
    delay = Float(title=u"Delay to wait tasks in queue (sec., 3 by default)", min=0.0, required=False, default=3.0)
    event = TextLine(title=u"IEvent utility name to stop workers.",
                     description=u"""Optional. If not defined event ill be created.
                                   Could be used for other tasks synchronization.""",
                     required=False)
            
class IInQueue(Interface):
    """ Input queue interface """
    in_queue = TextLine(title=u"IQueue utility name to be used as input tasks queue.", required=True) 
    
class IOutQueue(Interface):
    """ Output queue interface """
    out_queue = TextLine(title=u"IQueue utility name to be used as output tasks queue.", required=True)
    
class IInitStageDirective(IStageBase, IOutQueue):
    """ Initial stage. Requires output queue only to put new tasks. """
    
class IFinalStageDirective(IStageBase, IInQueue):
    """ Final stage. Requires input queue only to get tasks. """

class IStageDirective(IStageBase, IInQueue, IOutQueue):
    """ Regular stage. Requires both input and output queues. """      


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2008
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2008
#######################################################################

"""setup script class for the ZTK-based sterch.conveyor package

"""
__author__  = "Polscha Maxim (maxp@sterch.net)"
__license__ = "ZPL"

import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

def alltests():
    import logging
    import pkg_resources
    import unittest

    class NullHandler(logging.Handler):
        level = 50
        
        def emit(self, record):
            pass

    logging.getLogger().addHandler(NullHandler())

    suite = unittest.TestSuite()
    base = pkg_resources.working_set.find(
        pkg_resources.Requirement.parse('sterch.conveyor')).location
    for dirpath, dirnames, filenames in os.walk(base):
        if os.path.basename(dirpath) == 'tests':
            for filename in filenames:
                if filename.endswith('.py') and filename.startswith('test'):
                    mod = __import__(
                        _modname(dirpath, base, os.path.splitext(filename)[0]),
                        {}, {}, ['*'])
                    suite.addTest(mod.test_suite())
        elif 'tests.py' in filenames:
            continue
            mod = __import__(_modname(dirpath, base, 'tests'), {}, {}, ['*'])
            suite.addTest(mod.test_suite())
    return suite

setup( name='sterch.conveyor',
    version='0.1',
    url='http://pypi.sterch.net/sterch.conveyor',
    license='ZPL 2.1',
    description='Provides ZCML-based tools to define simultaneous tasks processing conveyors',
    author='Maksym Polscha',
    author_email='maxp@sterch.net',
    long_description=(
        read('README.txt')
        + '\n' +
        read('CHANGES.txt')
        ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Zope Public License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.5",
        "Programming Language :: Python :: 2.6",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],

    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['sterch'],
    include_package_data=True,
    exclude_package_data = { '': ['.gitignore', 'bootstrap.py', 'buildout.cfg'] },
    install_requires=['setuptools',
                        'zope.interface',
                        'zope.schema',
                        'zope.cachedescriptors',
                        'zope.configuration',
                        'zope.component',
                        'sterch.threading',
                        'sterch.queue',
                        'sterch.logfile',
                        ],
    extras_require={'test': ['zope.testing'],},                        
    test_suite='__main__.alltests',
    tests_require=['zope.testing'],
    )


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Help classes

"""
__author__  = "Polshcha Maxim (maxp@sterch.net)"
__license__ = "ZPL"

import traceback
import sys

from sterch.queue.interfaces import IQueue
from sterch.logfile.interfaces import ILog
from sterch.threading.interfaces import IEvent
from StringIO import StringIO
from zope.cachedescriptors.property import CachedProperty
from zope.component import getUtility, queryUtility

class OutQueueMixin(object):
    @CachedProperty
    def out_queue(self):
        if IQueue(self._out_queue, None): 
            return self._out_queue
        return getUtility(IQueue, name=self._out_queue)

class InQueueMixin(object):
    
    def has_tasks(self):
        """ Does queue have unfinished tasks. """
        return not self.in_queue.empty()
    
    def tasks_count(self):
        """ Return number of unfinished tasks. """
        return self.in_queue.qsize()

    @CachedProperty
    def in_queue(self):
        if IQueue(self._in_queue, None): 
            return self._in_queue
        return getUtility(IQueue, name=self._in_queue)

class EventMixin(object):
    @CachedProperty
    def event(self):
        if IEvent(self._event, None): 
            return self._event
        return getUtility(IEvent, name=self._event)
    
class LogMixin(object):
    @CachedProperty
    def log(self):
        log = queryUtility(ILog)
        class Log(object):
            def message(self, value): print value
        log = Log() if not log else log
        return log
    
    def traceback(self, ex):
        msg = "ERROR:" + str(ex)
        tb = StringIO()
        exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
        traceback.print_exception(exceptionType, exceptionValue, exceptionTraceback, file=tb)
        msg = msg + "\n" + tb.getvalue()
        if self.log: 
            self.log.message(msg)
        else:
            print msg


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Tests for IP address field
"""
__author__  = "Maxim Polscha (maxp@sterch.net)"
__license__ = "ZPL" 

import zcml

from activities import get_jobs_counter, reset_jobs, gl_N, get_results
try:
    from queue import Empty
except ImportError:
    from Queue import Empty
from setup import TestSetup
from sterch.conveyor.interfaces import IConveyor
from sterch.threading.interfaces import IEvent
from unittest import TestCase, makeSuite, main
from zope.component import queryUtility, getUtility, getUtilitiesFor
from zope.configuration.xmlconfig import XMLConfig

EXECUTION_TIME_LIMIT = 60

class Test(TestSetup):
    """Test the various zcml configurations"""
    
    def _clear_events(self):
        map(lambda n,e:e.clear(), getUtilitiesFor(IEvent))
    
    def _check_valid_config(self, config, uname, task_field='result'):
        XMLConfig(config, zcml)()
        c = queryUtility(IConveyor, name=uname)
        self.assertTrue(c is not None)
        reset_jobs()
        c.start()
        c.join(EXECUTION_TIME_LIMIT)
        self.assertFalse(c.isAlive())
        q = get_results()
        self.assertEqual(q.qsize(), gl_N)
        self.assertEqual(get_jobs_counter(), gl_N)
        try:
            while True:
                task = q.get(False)
                self.assertEqual(task['value'], task[task_field])
        except Empty:
            pass
 
    def test_correct_zcml_no_events(self):
        self._check_valid_config('valid_no_events.zcml', u"Test #1")

    def test_correct_zcml_with_events(self):
        self._clear_events()
        self._check_valid_config('valid_with_events.zcml', u"Test #2")
        # check events
        for name, event in getUtilitiesFor(IEvent):
            self.assertTrue(event.isSet())
            
    def test_2stages_only_no_events(self):
        self._check_valid_config('valid_2stages_no_events.zcml', u"Test #3", task_field='value')
    
    def test_2stages_only_with_events(self):
        self._clear_events()
        self._check_valid_config('valid_2stages_with_events.zcml', u"Test #4", task_field='value')
        for name in (u'Event #1', u'Event #2',):
            self.assertTrue(getUtility(IEvent, name=name).isSet())
        
def test_suite():
    suite = makeSuite(Test)
    return suite

if __name__ == '__main__':
    main(defaultTest='test_suite')


### -*- coding: utf-8 -*- #############################################
# Разработано компанией Стерх (http://sterch.net/)
# Все права защищены, 2010
#
# Developed by Sterch (http://sterch.net/)
# All right reserved, 2010
#######################################################################

""" Activities for test cass for sterch.conveyor
"""
__author__  = "Maxim Polscha (maxp@sterch.net)"
__license__ = "ZPL" 

from threading import Lock
try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty

gl_lock = Lock()
gl_jobs_counter = 0
gl_N = 50
gl_queue = Queue() 

def get_jobs_counter(): 
    global gl_jobs_counter
    return gl_jobs_counter

def reset_jobs(): 
    global gl_jobs_counter, gl_queue
    gl_jobs_counter = 0
    try:
        while True: gl_queue.get(False)
    except Empty:
        pass
    
def get_results(): 
    global gl_queue
    return gl_queue

def increment(job): 
    job['value'] +=1
    yield job
    
def mul10(job): 
    job['value'] *=10
    yield job

def neg(job): 
    job['value'] *= -1
    yield job

def jobs():
    """ jobs generator. Generates tasks and expected results for the tests """
    global gl_N
    for i in xrange(0, gl_N):
        yield dict(value = i, 
                   result = -(i * 10 + 1))
        
def check_results(job):
    """ Check value against expected result """
    global gl_jobs_counter, gl_N, gl_lock, gl_queue
    gl_queue.put(job)
    with gl_lock: gl_jobs_counter +=1
    assert gl_jobs_counter <= gl_N    